Stairs_docstring = """
Objects of class Stairs are used to represent a :ref:`step function <intro_tutorials.stepfunction>`.

See the :ref:`Stairs API <api.Stairs>` for details of attributes and methods.

The constructor provides the ability to combine a call to :meth:`Stairs.layer` which is useful for
creating Stairs objects via the split-apply-combine pattern facilitated by :meth:`pandas.DataFrame.groubpy`

Parameters
----------
frame : :class:`pandas.DataFrame`, optional
    A long-form collection of vectors that can be assigned to named variables
start : float, array-like of floats, str, default None
    Start point(s) of the interval(s).
    A value of None is interpreted as negative infinity.
end : float, array-like of floats, [description], by default Nonestr, default None
    End points(s) of the interval(s).
    A value of None is interpreted as positive infinity.
value : float, array-like of floats, str, default None
    Value(s) of the interval(s).
    A value of None is equivalent to a value of 1.
initial_value : float, default 0
    The value of the step function at negative infinity.
closed : {"left", "right"}
    Indicates whether the half-open intervals comprising the step function should be interpreted
    as left-closed or right-closed.

Returns
-------
:class:`Stairs`

See Also
--------
Stairs.layer

Examples
--------

.. plot::
    :context: close-figs

    >>> sc.Stairs(start=1, end=2)
    <staircase.Stairs, id=2375047689840>

    >>> sc.Stairs(start=1, end=2).to_frame()
    0  -inf    1      0
    1     1    2      1
    2     2  inf      0

    >>> sc.Stairs(start=1, end=2).plot()


.. plot::
    :context: close-figs

    >>> sc.Stairs(start=[0,1], end=[2,3]).to_frame()
    0  -inf    0      0
    1     0    1      1
    2     1    2      2
    3     2    3      1
    4     3  inf      0

    >>> sc.Stairs(start=[0,1], end=[2,3]).plot()


.. plot::
    :context: close-figs

    >>> data = pd.DataFrame(
    ...     {
    ...         "start":[0,1],
    ...         "end":[2,3],
    ...         "value":[3,4],
    ...     }
    ... )

    >>> sc.Stairs(frame=data, start="start", end="end", value="value").to_frame()
    0  -inf    0      0
    1     0    1      3
    2     1    2      7
    3     2    3      4
    4     3  inf      0

    >>> sc.Stairs(frame=data, start="start", end="end", value="value").plot()
    """



matplotlib_docstring = """
Plots the step function.

Parameters
----------
ax : :class:`matplotlib.axes.Axes`, optional
    Pre-existing axes to plot onto.
style : {"step", "hlines"}, default "step"
    Indicates whether the plot is based on :meth:`matplotlib.axes.Axes.step`
    or :meth:`matplotlib.axes.Axes.hlines`
arrows : bool, default False
    Whether or not to plot arrows where there exist intervals of infinite length
arrow_kwargs : dict or None, default None
    Additional keyword parameters passed to :meth:`matplotlib.axes.Axes.arrow`

Returns
-------
:class:`matplotlib.axes.Axes`

Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()


.. plot::
    :context: close-figs

    >>> s1.plot(arrows=True)


.. plot::
    :context: close-figs

    >>> s1.plot(style="hlines", arrows=True)

"""



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.plotting import register_matplotlib_converters

from staircase.util._decorators import Appender

from . import docstrings


def _draw_arrows(frame, ax, color, linewidth, **kwargs):

    ax.get_ylim()  # bugfix to trigger axes transform

    axis_to_data = ax.transAxes + ax.transData.inverted()
    axes_lims = axis_to_data.transform([(0, 0), (1, 1)])
    head_width = (axes_lims[1, 1] - axes_lims[0, 1]) * 0.03
    arrow_length = (axes_lims[1, 0] - axes_lims[0, 0]) * 0.1

    if len(frame) == 1:
        y = frame.iloc[0, -1]
        x = sum(ax.get_xlim()) / 2
        frame = pd.DataFrame(
            {
                "start": [None, x],
                "end": [x, None],
                "value": [y, y],
            }
        )
    kwargs = {
        "head_width": head_width,
        "head_length": arrow_length * 0.5,
        "color": color,
        "linewidth": linewidth,
        **kwargs,
    }
    if not pd.isnull(frame.iloc[-1]["start"]):
        x = frame.iloc[-1]["start"]
        y = frame.iloc[-1]["value"]
        ax.arrow(
            x,
            y,
            arrow_length,
            0,
            **kwargs,
        )
    if not pd.isnull(frame.iloc[0]["value"]):
        x = frame.iloc[0]["end"]
        y = frame.iloc[0]["value"]
        ax.arrow(
            x,
            y,
            -arrow_length,
            0,
            **kwargs,
        )


def _plot_matplotlib_steps(frame, ax, **kwargs):
    frame.loc[0, "start"] = frame.loc[0, "end"]

    return ax.step(
        frame["start"],
        frame["value"],
        where="post",
        **kwargs,
    )


def _plot_matplotlib_hlines(frame, ax, **kwargs):
    plot_data = frame.iloc[1:-1].loc[lambda df: df["value"].notnull()]

    return ax.hlines(
        plot_data["value"],
        plot_data["start"],
        plot_data["end"],
        **kwargs,
    )


@Appender(docstrings.matplotlib_docstring, join="\n", indents=1)
def plot(
    self, ax=None, style="step", arrows=False, linewidth=1, arrow_kwargs=None, **kwargs
):
    assert style in ("step", "hlines")
    register_matplotlib_converters()
    # TODO warn if style is hlines and arrow is None
    if arrow_kwargs is None:
        arrow_kwargs = {}
    if ax is None:
        _, ax = plt.subplots()
    frame = self.to_frame()
    if len(frame) == 1:
        value = frame.iloc[0, -1]
        if not np.isnan(value):
            color = kwargs.get("color", None)
            if arrows:
                _draw_arrows(frame, ax, color, linewidth, **arrow_kwargs)
            else:
                ax.axhline(value, color=color, **kwargs)
        return ax
    if style == "step":
        lines = _plot_matplotlib_steps(frame, ax=ax, linewidth=linewidth, **kwargs)
        color = lines[0].get_color()
    else:
        lines = _plot_matplotlib_hlines(frame, ax=ax, linewidth=linewidth, **kwargs)
        color = lines.get_colors()[0]
    if arrows or style == "hlines":
        _draw_arrows(frame, ax, color, linewidth, **arrow_kwargs)
    return ax



# PROPERTIES -------------------------------------

_stat_docstring = """
The {calc} of the step function.

Returns
-------
float
     The {calc} of the step function

See Also
--------
{see_also}

{example}
"""

_stat_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> {stairs1}.plot(arrows=True)
    >>> {stairs1}.{func}()
    {result1}

.. plot::
    :context: close-figs

    >>> {stairs2}.plot(arrows=True)
    >>> {stairs2}.{func}()
    {result2}
"""

simple_min_example = _stat_example.format(
    func="min",
    stairs1="s1",
    result1="-1",
    stairs2="s3.clip(2,4)",
    result2="0.0",
)
simple_max_example = _stat_example.format(
    func="max",
    stairs1="s1",
    result1="1",
    stairs2="s3.clip(2,4)",
    result2="1.0",
)
integral_example = _stat_example.format(
    func="integral",
    stairs1="s1",
    result1="1",
    stairs2="s3",
    result2="0.5",
)
mean_example = _stat_example.format(
    func="mean",
    stairs1="s1",
    result1="0.25",
    stairs2="s3",
    result2="0.166666",
)
mode_example = _stat_example.format(
    func="mode", stairs1="s1", result1="1", stairs2="s4", result2="0.5"
)
median_example = _stat_example.format(
    func="median",
    stairs1="s1",
    result1="0.5",
    stairs2="s4",
    result2="0.0",
)
std_example = _stat_example.format(
    func="std",
    stairs1="s1",
    result1="0.829156",
    stairs2="s3",
    result2="0.897527",
)
var_example = _stat_example.format(
    func="var",
    stairs1="s1",
    result1="0.6875",
    stairs2="s3",
    result2="0.805555",
)


def _get_example(calculation):

    return {
        "mean": mean_example,
        "integral": integral_example,
        # "percentile": percentile_prop_example,
        "median": median_example,
        "mode": mode_example,
        "var": var_example,
        "std": std_example,
    }[calculation]


_calc_map = {
    "integral": "integral",
    "percentile": "x-th percentile",
    "var": "variance",
    "std": "standard deviation",
}

_see_also_map = {
    "mean": "Stairs.integral, Stairs.rolling_mean",
    "integral": "Stairs.mean",
    "percentile": "Stairs.median, Stairs.fractile",
    "median": "Stairs.percentile, Stairs.fractile",
    "mode": "Stairs.mean, Stairs.median",
    "var": "Stairs.std",
    "std": "Stairs.var",
}


def _gen_docstring(calculation):

    calc = _calc_map.get(calculation, calculation)
    see_also = _see_also_map[calculation]
    example = _get_example(calculation)

    if property:
        docstring = _stat_docstring

    return docstring.format(
        calc=calc,
        see_also=see_also,
        example=example,
    )


mean_docstring = _gen_docstring("mean")
integral_docstring = _gen_docstring("integral")
median_docstring = _gen_docstring("median")
mode_docstring = _gen_docstring("mode")
var_docstring = _gen_docstring("var")
std_docstring = _gen_docstring("std")


# AGG ------------------------------------------------------------------

agg_docstring = """
Returns the aggregation of the step function defined over a domain.

Parameters
----------
name : {'max', 'min', 'mode', 'median', 'mean', 'integral', 'var', 'std'}
    The name of the function which which to perform the aggregation.
where : tuple or list of length two, optional
    Indicates the domain interval over which to evaluate the step function.
    Default is (-sc.inf, sc.inf) or equivalently (None, None).
closed : {'both', 'left', 'right', 'neither'}, optional
    Note that this parameter will only change the result for 'min' or 'max' aggregations.
    Indicates whether the interval defined by *where* is closed on both sides, left-closed,
    right-closed, or neither.  Default value is taken from the *closed* attribute of the step function.
    The default values will be used if *closed=None* is used.

Returns
-------
float

See Also
--------
Stairs.max, Stairs.min, Stairs.mode, Stairs.median, Stairs.mean, Stairs.integral, Stairs.var, Stairs.std

Examples
--------
.. plot::
    :context: close-figs

    >>> s1.plot(arrows=True)

>>> s1.agg(where=(1,4), name="min")
0

>>> s1.agg(where=(1,4), name="min", closed="both")
-1

>>> s1.agg(where=(1,4), name="integral")
2.0
"""


# VALUE SUMS -----------------------------------------------------

value_sums_docstring = """
Characterises the distribution of step function values with a series of their sums.

The result is a :class:`pandas.Series`, indexed by the values of a the step function.
The values of the series are the lengths of the intervals whose value is equal
to the corresponding index.

This method is analagous to :func:`pandas.Series.value_counts` over a continous domain.

Parameters
----------
dropna : bool, default True
    Indicates whether to include intervals where the step function is undefined
group : bool, default True
    Indicates whether to group the index values together into a unique index.

Returns
-------
:class:`pandas.Series`

Examples
--------
.. plot::
    :context: close-figs

    >>> s1.plot()

>>> s1.value_sums()
-1    1
 0    1
 1    2
dtype: int64

.. plot::
    :context: close-figs

    >>> s3.plot()

    >>> s3.value_sums(dropna=False)
    -1.0    1.0
     0.0    0.5
     1.0    1.5
     NaN    1.0
    dtype: float64
"""


# MIN MAX VALUES -------------------------------------------------

_min_max_values_in_range_docstring = """
Returns the {calc} of the step function defined over a domain.

Parameters
----------
where : tuple or list of length two, optional
    Indicates the domain interval over which to evaluate the step function.
    Default is (-sc.inf, sc.inf) or equivalently (None, None).
closed: {{'both', 'left', 'right', 'neither'}}, optional
    Indicates whether the interval defined by *where* is closed on both sides, left-closed,
    right-closed, or neither.  Default value is taken from the *closed* attribute of the step function.
    The default values will be used if *closed=None* is used.

Returns
-------
{return_val}

See Also
--------
{see_also}

{example}
"""

values_in_range_example = """
Examples
--------
.. plot::
    :context: close-figs

    >>> s1.plot(arrows=True)

>>> s1.values_in_range()
array([-1,  0,  1], dtype=int64)

>>> s1.values_in_range((1,4))
array([0,  1], dtype=int64)

>>> s1.values_in_range((1,4), closed='both')
array([-1,  0,  1], dtype=int64)
"""

min_example = """
Examples
--------
.. plot::
    :context: close-figs

    >>> from staircase.api import min
    >>> s1.plot(arrows=True)

>>> min(s1)
-1

>>> min(s1, (1,4))
0

>>> min(s1, where=(1,4), closed='both')
-1
"""

max_example = """
Examples
--------
.. plot::
    :context: close-figs

    >>> from staircase.stairs import max
    >>> s1.plot(arrows=True)

>>> max(s1)
-1

>>> max(s1, (4,5))
0

>>> max(s1, where=(4,5), closed='both')
0
"""


values_in_range_docstring = _min_max_values_in_range_docstring.format(
    calc="range (i.e. values)",
    return_val=r":class:`numpy.ndarray`\n    A numpy array of unique values",
    see_also="Stairs.min, Stairs.max",
    example=values_in_range_example,
)


min_docstring = _min_max_values_in_range_docstring.format(
    calc="minimum",
    return_val="float",
    see_also="Stairs.min, staircase.stairs.max",
    example=min_example,
)


max_docstring = _min_max_values_in_range_docstring.format(
    calc="minimum",
    return_val="float",
    see_also="Stairs.max, staircase.stairs.min",
    example=max_example,
)


hist_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot(arrows=True)

>>> s1.hist()
[-1, 0)    1.0
[0, 1)     1.0
[1, 2)     2.0
dtype: float64

>>> s1.hist(closed="right")
(-2, -1]    1.0
(-1, 0]     1.0
(0, 1]      2.0
dtype: float64

>>> s1.hist(bins=[-1,1,2])
[-1, 1)    2.0
[1, 2)     2.0
dtype: float64

>>> s1.hist(bins=[-1,1,2], stat="frequency")
[-1, 1)    1.0
[1, 2)     2.0
dtype: float64

>>> s1.hist(bins=[-1,1,2], stat="density")
[-1, 1)    0.333333
[1, 2)     0.333333
dtype: float64

>>> s1.hist(bins=[-1,1,2], stat="probability")
[-1, 1)    0.5
[1, 2)     0.5
dtype: float64
"""

ecdf_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s2.plot(arrows=True)

.. plot::
    :context: close-figs

    >>> s2.ecdf
    <staircase.ECDF, id=1872531189136>

    >>> s2.ecdf.plot()
    >>> plt.show()

>>> print(f'{ecdf(0):.2%} of values for s2 are less than or equal to 0')
75.00% of values, for s2 between 1 and 5, are less than or equal to 0

>>> print(f'{ecdf(0, how="left"):.2%} of values for s2 are strictly less than 0')
50.00% of values, for s2 between 1 and 5, are strictly less than 0

>>> print(f'{ecdf(0.2) - ecdf(-1):.2%} of values for s2 are in (-1, 0.2]')
25.00% of values, for s2 between 1 and 5, are in (-1, 0.2]

>>> print(f'{ecdf(0.2, how="left") - ecdf(-1, how="left"):.2%} of values for s2 are in [-1, 0.2)')
75.00% of values, for s2 between 1 and 5, are in [-1, 0.2)
"""

# ecdf_stairs_example = """
# Examples
# --------

# .. plot::
#     :context: close-figs

#     >>> s2.plot()

# .. plot::
#     :context: close-figs

#     >>> s2.ecdf
#     <staircase.ECDF, id=1872531189136>

#     >>> s2.ecdf.plot()
#     >>> plt.show()


#     >>> print(f'{ecdf(0):.2%} of values, for s2 between 1 and 5, are less than or equal to 0')
#     75.00% of values, for s2 between 1 and 5, are less than or equal to 0

#     >>> print(f'{ecdf(0, how="left"):.2%} of values, for s2 between 1 and 5, are strictly less than 0')
#     50.00% of values, for s2 between 1 and 5, are strictly less than 0

#     >>> print(f'{ecdf(0.2) - ecdf(-1):.2%} of values, for s2 between 1 and 5, are in (-1, 0.2]')
#     25.00% of values, for s2 between 1 and 5, are in (-1, 0.2]

#     >>> print(f'{ecdf(0.2, how="left") - ecdf(-1, how="left"):.2%} of values, for s2 between 1 and 5, are in [-1, 0.2)')
#     75.00% of values, for s2 between 1 and 5, are in [-1, 0.2)
# """


# _prop_docstring = """
# The {calc} of the step function values.

# Parameters
# ----------


# Returns
# -------
# float
#      The {calc} of the step function

# See Also
# --------
# {see_also}

# {example}
# """

# percentile_prop_example = prop_example.format(
#     func1="percentile",
#     stairs1="s1",
#     result1="<staircase.Percentiles, id=1872507658000>",
#     stairs2="s2",
#     result2="-1.0",
# )


# func_example = """
# Examples
# --------

# .. plot::
#     :context: close-figs

#     >>> {stairs1}.plot()
#     >>> {stairs1}.{func}(where=(2, 4.5))
#     {result1}

#     >>> {stairs2}.plot()
#     >>> {stairs2}.{func}(where=(None, 4.5))
#     {result2}
# """

# integral_func_example = func_example.format(
#     func="integral", stairs1="s1", result1="1", stairs2="s3", result2="0.5",
# )
# mean_func_example = func_example.format(
#     func="mean", stairs1="s1", result1="0.25", stairs2="s3", result2="0.166666",
# )
# mode_func_example = func_example.format(
#     func="mode", stairs1="s1", result1="1", stairs2="s4", result2="0.5"
# )
# median_func_example = func_example.format(
#     func="median", stairs1="s1", result1="0.5", stairs2="s4", result2="0.0",
# )
# std_func_example = func_example.format(
#     func="std", stairs1="s1", result1="0.829156", stairs2="s3", result2="0.897527",
# )
# var_func_example = func_example.format(
#     func="var", stairs1="s1", result1="0.6875", stairs2="s3", result2="0.805555",
# )
# percentile_func_example = func_example.format(
#     func1="percentile",
#     stairs1="s1",
#     result1="<staircase.Percentiles, id=1872507658000>",
#     stairs2="s2",
#     result2="-1.0",
# )



import math

import numpy as np
import pandas as pd

import staircase as sc


class Xtiles(sc.core.stairs.Stairs):

    class_name = None
    scale_factor = None

    def sample(self, x):
        return (self.limit(x, side="left") + self.limit(x, side="right")) / 2

    def __call__(self, *args, **kwargs):
        return self.sample(*args, **kwargs)

    @classmethod
    def from_ecdf(cls, ecdf):
        assert ecdf._data is not None
        return cls._new(
            initial_value=ecdf._data.index[0],
            data=pd.DataFrame(
                {"value": np.append(ecdf._data.index, ecdf._data.index[-1])},
                index=np.append(0, ecdf._get_values().values * cls.scale_factor),
            ),
        )


class Percentiles(Xtiles):

    class_name = "Percentiles"
    scale_factor = 100


class Fractiles(Xtiles):

    class_name = "Fractiles"
    scale_factor = 1

    def to_percentiles(self):

        data = self._get_values().copy()
        data.index = data.index * 100
        return Percentiles._new(initial_value=self.initial_value, data=data.to_frame())


class ECDF(sc.core.stairs.Stairs):

    class_name = "ECDF"

    @staticmethod
    def from_stairs(stairs):
        ecdf_deltas = stairs.value_sums(dropna=True).rename("delta")
        deltas_sum = ecdf_deltas.sum()
        normalized_probability_deltas = ecdf_deltas / deltas_sum

        ecdf = ECDF._new(
            initial_value=0,
            data=normalized_probability_deltas.to_frame(),
            closed="left",
        )
        ecdf._denormalize_probability_factor = deltas_sum
        return ecdf

    def hist(self, bins="unit", closed="left", stat="sum"):

        step_points = self.step_points
        if isinstance(bins, str) and bins == "unit":
            round_func = math.floor if closed == "left" else math.ceil
            bins = range(
                round_func(min(step_points)) - (closed == "right"),
                round_func(max(step_points)) + (closed == "left") + 1,
            )

        if not isinstance(bins, pd.IntervalIndex):
            bins = pd.IntervalIndex.from_breaks(bins, closed=closed)

        values = self.limit(bins.right, side=bins.closed) - self.limit(
            bins.left, side=bins.closed
        )

        # inspired by seaborn.histplot
        if stat != "probability":
            values = values * self._denormalize_probability_factor
            if stat in ("frequency", "density"):
                widths = bins.map(lambda i: i.right - i.left)
                if stat == "frequency":
                    values = values / widths
                elif stat == "density":
                    values = values / np.dot(values, widths)

        return pd.Series(
            data=values,
            index=bins,
            # dtype="float64",
        )

    def percentiles(self):
        return Percentiles.from_ecdf(self)

    def fractiles(self):
        return Fractiles.from_ecdf(self)


class Dist:
    def __init__(self, stairs):
        self._reset()
        self._stairs = stairs

    def _reset(self):
        self._ecdf = None
        self._fractiles = None
        self._percentiles = None

    @property
    def ecdf(self):
        if self._ecdf is None:
            self._ecdf = ECDF.from_stairs(self._stairs)
        return self._ecdf

    def hist(self, bins="unit", closed="left", stat="sum"):
        return self.ecdf.hist(bins, closed, stat)

    @property
    def fractile(self):
        if self._fractiles is None:
            self._fractiles = Fractiles.from_ecdf(self.ecdf)
        return self._fractiles

    @property
    def percentile(self):
        if self._percentiles is None:
            self._percentiles = Percentiles.from_ecdf(self.ecdf)
        return self._percentiles

    def quantiles(self, n):
        assert n > 0 and isinstance(n, int)
        return self.fractile(np.linspace(0, 1, n + 1)[1:-1])


# def _get_dist(self):
#     if self._dist is None:
#         self._dist = Dist(self)
#     return self._dist



import re

_x_agg_docstring = """
Returns a :class:`staircase.Stairs` object representing the {calc} of multiple step functions.

{params}

Returns
-------
:class:`Stairs`

See Also
--------
{see_also}

Examples
--------
{example}
"""

_toplevel_params = """
Parameters
----------
collection : tuple, list, numpy array, dict or pandas.Series
    The Stairs instances to aggregate"""


def _make_see_also(args):
    if args is None:
        return ""
    else:
        return f"See Also\n--------\n{args}"


_example = """
>>> import staircase as sc
>>> {arr_def}

.. plot::
    :context: close-figs
    :include-source: False

    >>> {arr_def}
    >>> stair_list = [{plots}]
    >>> fig, axes = plt.subplots(nrows=1, ncols={ncols},  figsize=({width},3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, {plot_titles}, stair_list):
    ...     stair_instance.plot(ax=ax, arrows=True)
    ...     ax.set_title(title)
"""


def _get_plot_titles(which, method):
    if method == "agg":
        result = (
            f"sc.{method}(arr, np.std)"
            if which == "toplevel"
            else f"arr.{method}(np.std)"
        )
    else:
        result = f"sc.{method}(arr)" if which == "toplevel" else f"arr.{method}()"
    return ["s1", "s2", result]


def _gen_example(plot_titles, arr_def):
    # plot_titles is a list of strings
    ncols = len(plot_titles)
    plots = ", ".join(plot_titles)
    return _example.format(
        arr_def=arr_def,
        plots=plots,
        ncols=ncols,
        plot_titles=plot_titles,
        width=ncols + 5,
    )


def make_docstring(which, method):

    if method in ("logical_or", "logical_and"):
        return _make_logical_docstring(which, method)

    if method in ("limit"):
        return _make_limit_docstring(which)

    if method in ("sample"):
        return _make_sample_docstring(which)

    if method in ("cov", "corr"):
        return _make_cov_corr_docstring(which, method)

    if method in ("plot"):
        return _make_plot_docstring(which)

    if which == "array":
        see_also = f":func:`staircase.{method}`, :meth:`pandas.Series.{method}`"
        params = ""
        arr_def = "arr = sc.StairsArray([s1,s2])"
    elif which == "toplevel":
        see_also = (
            f":meth:`staircase.StairsArray.{method}`, :meth:`pandas.Series.{method}`"
        )
        params = _toplevel_params.format(calc=method)
        arr_def = "arr = [s1, s2]"

    example = _gen_example(
        _get_plot_titles(which, method),
        arr_def=arr_def,
    )
    if method == "agg":
        example = ">>> import numpy as np\n" + example

    return re.sub(
        "\n\n+",
        "\n\n",
        _x_agg_docstring.format(
            calc={"min": "minimum", "max": "maximum", "agg": "aggregate"}.get(
                method, method
            ),
            params=params,
            see_also=see_also,
            example=example,
        ),
    )


_logical_base_docstring = """
Returns a boolean-valued step function resulting from a {logical_method} operation.

The operands are the Stairs instances in the object the accessor belongs to.

Returns
-------
:class:`staircase.Stairs`
    boolean-valued step function

Examples
--------

>>> {setup}

.. plot::
    :context: close-figs
    :include-source: False

    >>> {setup}
    >>> stair_list = [s1, s2, {result}]
    >>> fig, axes = plt.subplots(nrows=1, ncols=3,  figsize=(8,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ["s1", "s2", "{result}"], stair_list):
    ...     stair_instance.plot(ax=ax, arrows=True)
    ...     ax.set_title(title)
"""


def _make_logical_docstring(which, method):
    if which == "accessor":
        setup = 'arr = pd.Series([s1, s2], dtype="Stairs")'
        result = f"arr.sc.{method}()"
    elif which == "toplevel":
        setup = "arr = [s1, s2]"
        result = f"sc.{method}(arr)"
    elif which == "array":
        setup = "arr = sc.StairsArray([s1, s2])"
        result = f"arr.{method}()"

    logical_method = {
        "logical_or": "logical-or",
        "logical_and": "logical-and",
    }[method]
    return _logical_base_docstring.format(
        logical_method=logical_method, setup=setup, result=result
    )


_limit_base = """
Takes a collection of Stairs instances and evaluates their limits across a set of points.

Technically the results of this function should be considered as :math:`\\lim_{{x \\to z^{{-}}}} f(x)`
or :math:`\\lim_{{x \\to z^{{+}}}} f(x)`, when *side* = 'left' or *side* = 'right' respectively. See
:ref:`A note on interval endpoints<getting_started.interval_endpoints>` for an explanation.

Parameters
{collection_param}
x : scalar or vector data
    The points at which to sample the Stairs instances.  Must belong to the step function domain.
side : {{'left', 'right'}}, default 'right'
    if points where step changes occur do not coincide with x then this parameter
    has no effect.  Where a step changes occurs at a point given by x, this parameter
    determines if the step function is evaluated at the interval to the left, or the right.

Returns
-------
:class:`pandas.DataFrame`
    {return_desc}

See Also
--------
{see_also}

Examples
--------

.. plot::
    :context: close-figs
    :include-source: False

    >>> stair_list = [s2, s3]
    >>> fig, axes = plt.subplots(nrows=1, ncols=2,  figsize=(7,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ["s2", "s3"], stair_list):
    ...     stair_instance.plot(ax=ax, arrows=True)
    ...     ax.set_title(title)

>>> import staircase as sc
>>> {setup}
>>> [s2.closed, s3.closed]
["left", "left]

>>> {calc_preamble}[2,3,4], side="left"))
     2    3    4
0  0.5  0.0 -1.0
1  1.0  NaN  1.0

>>> {calc_preamble}[2,3,4], side="right"))
     2    3    4
0  0.0 -1.0 -1.0
1  0.0  NaN -1.0
"""

_collection_param = """----------
collection : array-like, dictionary or pandas.Series
    The Stairs instances at which to evaluate"""

_top_level_limit_extra_examples = """
>>> stairs = {"s2":s2, "s3":s3}
>>> sc.limit(stairs, [2,3,4], side="left"))
      2    3    4
s1  0.5  0.0 -1.0
s2  1.0  NaN  1.0

>>> index=pd.MultiIndex.from_tuples([("a", "s2"), ("b", "s3")])
>>> stairs = pd.Series([s2,s3], index=index)
>>> sc.limit(stairs, [2,3,4], side="left"))
        2    3    4
a s1  0.5  0.0 -1.0
b s2  1.0  NaN  1.0
"""


def _make_limit_docstring(which):
    if which == "toplevel":
        collection_param = _collection_param
        setup = "stairs = [s2, s3]"
        calc_preamble = "sc.limit(stairs, "
        return_desc = """A dataframe, where rows correspond to the Stairs instances in *collection*.
    and columns correspond to the points in *x*.  If *collection* is a dictionary then the
    resulting dataframe will be indexed by the dictionary keys.  If *collection* is a
    :class:`pandas.Series` then the dataframe will have the same index as the series."""
        see_also = ":meth:`staircase.core.arrays.accessor.StairsAccessor.limit`, :meth:`staircase.StairsArray.limit`, :meth:`staircase.sample`"
    elif which == "accessor":
        collection_param = "----------"
        setup = 'stairs = pd.Series([s2, s3], dtype="Stairs")'
        calc_preamble = "stairs.sc.limit("
        return_desc = """A dataframe, where rows correspond to the Stairs instances in the :class:`pandas.Series`.
    and columns correspond to the points in *x*.  The dataframe will have the same index as the Series."""
        see_also = ":meth:`staircase.limit`, :meth:`staircase.StairsArray.limit`, :meth:`staircase.core.arrays.accessor.StairsAccessor.sample`"
    elif which == "array":
        collection_param = "----------"
        setup = "stairs = sc.StairsArray([s2, s3])"
        calc_preamble = "stairs.limit("
        return_desc = """A dataframe, where rows correspond to the Stairs instances in the :class:`StairsArray`.
    and columns correspond to the points in *x*."""
        see_also = ":meth:`staircase.limit`, :meth:`staircase.core.arrays.accessor.StairsAccessor.limit`, :meth:`staircase.StairsArray.sample`"
    doc = _limit_base.format(
        collection_param=collection_param,
        setup=setup,
        calc_preamble=calc_preamble,
        return_desc=return_desc,
        see_also=see_also,
    )
    if which == "toplevel":
        doc += _top_level_limit_extra_examples
    return doc


_sample_base = """
Takes a collection of Stairs instances and evaluates their values across a set of points.

Parameters
{collection_param}
x : scalar or vector data
    The points at which to sample the Stairs instances.  Must belong to the step function domain.

Returns
-------
:class:`pandas.DataFrame`
    {return_desc}

See Also
--------
{see_also}

Examples
--------

.. plot::
    :context: close-figs
    :include-source: False

    >>> stair_list = [s2, s3]
    >>> fig, axes = plt.subplots(nrows=1, ncols=2,  figsize=(7,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ["s2", "s3"], stair_list):
    ...     stair_instance.plot(ax=ax, arrows=True)
    ...     ax.set_title(title)

>>> import staircase as sc
>>> {setup}
>>> [s2.closed, s3.closed]
["left", "left]

>>> {calc_preamble}[2,3,4]))
     2    3    4
0  0.0 -1.0 -1.0
1  0.0  NaN -1.0
"""

_top_level_sample_extra_examples = """
>>> stairs = {"s2":s2, "s3":s3}
>>> sc.sample(stairs, [2,3,4]))
      2    3    4
s1  0.0 -1.0 -1.0
s2  0.0  NaN -1.0

>>> index=pd.MultiIndex.from_tuples([("a", "s2"), ("b", "s3")])
>>> stairs = pd.Series([s2,s3], index=index)
>>> sc.sample(stairs, [2,3,4], side="left"))
        2    3    4
a s1  0.0 -1.0 -1.0
b s2  0.0  NaN -1.0
"""


def _make_sample_docstring(which):
    if which == "toplevel":
        collection_param = _collection_param
        setup = "stairs = [s2, s3]"
        calc_preamble = "sc.sample(stairs, "
        return_desc = """A dataframe, where rows correspond to the Stairs instances in *collection*.
    and columns correspond to the points in *x*.  If *collection* is a dictionary then the
    resulting dataframe will be indexed by the dictionary keys.  If *collection* is a
    :class:`pandas.Series` then the dataframe will have the same index as the series."""
        see_also = ":meth:`staircase.core.arrays.accessor.StairsAccessor.sample`, :meth:`staircase.StairsArray.sample`, :meth:`staircase.limit`"
    elif which == "accessor":
        collection_param = "----------"
        setup = 'stairs = pd.Series([s2, s3], dtype="Stairs")'
        calc_preamble = "stairs.sc.sample("
        return_desc = """A dataframe, where rows correspond to the Stairs instances in the :class:`pandas.Series`.
    and columns correspond to the points in *x*.  The dataframe will have the same index as the Series."""
        see_also = ":meth:`staircase.sample`, :meth:`staircase.StairsArray.sample`, :meth:`staircase.core.arrays.accessor.StairsAccessor.limit`"
    elif which == "array":
        collection_param = "----------"
        setup = "stairs = sc.StairsArray([s2, s3])"
        calc_preamble = "stairs.sample("
        return_desc = """A dataframe, where rows correspond to the Stairs instances in the :class:`StairsArray`.
    and columns correspond to the points in *x*."""
        see_also = ":meth:`staircase.sample`, :meth:`staircase.core.arrays.accessor.StairsAccessor.sample`, :meth:`staircase.StairsArray.limit`"
    doc = _sample_base.format(
        collection_param=collection_param,
        setup=setup,
        calc_preamble=calc_preamble,
        return_desc=return_desc,
        see_also=see_also,
    )
    if which == "toplevel":
        doc += _top_level_sample_extra_examples
    return doc


_cov_corr_base = """
Calculates the {calc_name} matrix for a collection of :class:`Stairs` instances

Parameters
{collection_param}
lower : int, float or pandas.Timestamp
    lower bound of the interval on which to perform the calculation
upper : int, float or pandas.Timestamp
    upper bound of the interval on which to perform the calculation

Returns
-------
:class:`pandas.DataFrame`
    The {calc_name} matrix

See Also
--------
{see_also}

Examples
--------
{examples}
"""

_cov_examples = """
>>> import staircase as sc
>>> {setup}

>>> {calc_method}
          0          1          2
0  0.687500   0.140496   0.652893
1  0.140496   0.471074   0.611570
2  0.652893   0.611570   1.264463
"""

_corr_examples = """
>>> import staircase as sc
>>> {setup}
>>> {calc_method}
          0          1          2
0  1.000000   0.246878   0.700249
1  0.246878   1.000000   0.792407
2  0.700249   0.792407   1.000000
"""

_top_level_cov_extra_examples = """
>>> stairs = {"s1":s1, "s2":s2, "s1+s2":s1+s2}
>>> sc.cov(stairs, [2,3,4]))
              s1         s2      s1+s2
s1      0.687500   0.140496   0.652893
s2      0.140496   0.471074   0.611570
s1+ s2  0.652893   0.611570   1.264463
"""


_top_level_corr_extra_examples = """
>>> stairs = {"s1":s1, "s2":s2, "s1+s2":s1+s2}
>>> sc.corr(stairs, [2,3,4]))
              s1         s2      s1+s2
s1      1.000000   0.246878   0.700249
s2      0.246878   1.000000   0.792407
s1+ s2  0.700249   0.792407   1.000000
"""


def _make_cov_corr_docstring(which, method):
    other = {"cov": "corr", "corr": "cov"}[method]
    calc_name = {"cov": "covariance", "corr": "correlation"}[method]
    examples = {"cov": _cov_examples, "corr": _corr_examples}[method]
    if which == "toplevel":
        collection_param = _collection_param
        setup = "stairs = [s1, s2, s1+s2]"
        calc_method = f"sc.{method}(stairs)"
        see_also = f":meth:`staircase.core.arrays.accessor.StairsAccessor.{method}`, :meth:`staircase.StairsArray.{method}`, :meth:`staircase.{other}`"
    elif which == "accessor":
        collection_param = "----------"
        setup = 'stairs = pd.Series([s1, s2, s1+s2], dtype="Stairs")'
        calc_method = f"stairs.sc.{method}()"
        see_also = f":meth:`staircase.{method}`, :meth:`staircase.StairsArray.{method}`, :meth:`staircase.core.arrays.accessor.StairsAccessor.{other}`"
    elif which == "array":
        collection_param = "----------"
        setup = "stairs = sc.StairsArray([s1, s2, s1+s2])"
        calc_method = f"stairs.{method}()"
        see_also = f":meth:`staircase.{method}`, :meth:`staircase.core.arrays.accessor.StairsAccessor.{method}`, :meth:`staircase.StairsArray.{other}`"
    doc = _cov_corr_base.format(
        collection_param=collection_param,
        calc_name=calc_name,
        see_also=see_also,
        examples=examples.format(setup=setup, calc_method=calc_method),
    )
    if which == "toplevel":
        if method == "cov":
            doc += _top_level_cov_extra_examples
        else:
            doc += _top_level_corr_extra_examples
    return doc


_plot_base = """
Plots a collection of :class:`Stairs` instances to a single axes

Parameters
{collection_param}
ax : :class:`matplotlib.axes.Axes`, optional
    If supplied will plot to this axes
labels : array-like, optional
    Used to label each step function.  Use `ax.legend()` to show.
**kwargs :
    Additional parameters are the same as those for :meth:`staircase.Stairs.plot`.

Returns
-------
:class:`matplotlib.axes.Axes`

See Also
--------
{see_also}

Examples
--------

>>> import staircase as sc
>>> import matplotlib.pyplot as plt

.. plot::
    :context: close-figs

    >>> {setup}
    >>> {calc_method}
    >>> ax.legend()
"""

_top_level_plot_extra_examples = """
.. plot::
    :context: close-figs

    >>> stairs = {"s1":s1, "s2":s2}
    >>> _, ax = plt.subplots(figsize=(4,3))
    >>> sc.plot(stairs, ax)
    >>> ax.legend()
"""

_accessor_plot_extra_examples = """
.. plot::
    :context: close-figs

    >>> stairs = pd.Series([s2, s3], index=["s2", "s3"], dtype="Stairs")
    >>> _, ax = plt.subplots(figsize=(4,3))
    >>> stairs.sc.plot(ax)
    >>> ax.legend()
"""

_array_plot_extra_examples = """
.. plot::
    :context: close-figs

    >>> stairs = sc.StairsArray([s2, s3])
    >>> _, ax = plt.subplots(figsize=(4,3))
    >>> stairs.plot(ax, labels=("s2", "s3"))
    >>> ax.legend()
"""


def _make_plot_docstring(which):
    if which == "toplevel":
        collection_param = _collection_param
        setup = "stairs = [s2, s3]\n    >>> _, ax = plt.subplots(figsize=(4,3))"
        calc_method = "sc.plot(stairs, ax), "
        see_also = ":meth:`staircase.core.arrays.accessor.StairsAccessor.plot`, :meth:`staircase.StairsArray.plot`, :meth:`staircase.Stairs.plot`"
    elif which == "accessor":
        collection_param = "----------"
        setup = 'stairs = pd.Series([s2, s3], dtype="Stairs")\n    >>> _, ax = plt.subplots(figsize=(4,3))'
        calc_method = "stairs.sc.plot(ax)"
        see_also = ":meth:`staircase.plot`, :meth:`staircase.StairsArray.plot`, :meth:`staircase.Stairs.plot`"
    elif which == "array":
        collection_param = "----------"
        setup = "stairs = sc.StairsArray([s2, s3])\n    >>> _, ax = plt.subplots(figsize=(4,3))"
        calc_method = "stairs.plot(ax)"
        see_also = ":meth:`staircase.plot`, :meth:`staircase.core.arrays.accessor.StairsAccessor.plot`, :meth:`staircase.Stairs.plot`"
    doc = _plot_base.format(
        collection_param=collection_param,
        setup=setup,
        calc_method=calc_method,
        see_also=see_also,
    )
    if which == "toplevel":
        doc += _top_level_plot_extra_examples
    elif which == "accessor":
        doc += _accessor_plot_extra_examples
    return doc


_extension_array_setup = """
>>> {var} = sc.StairsArray({stair_list})

.. plot::
    :context: close-figs
    :include-source: False

    >>> stair_list = {stair_list}
    >>> fig, axes = plt.subplots(nrows=1, ncols=2,  figsize=(7,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ["{var}[0]", "{var}[1]"], stair_list):
    ...     stair_instance.plot(ax=ax, arrows=True)
    ...     ax.set_title(title)
"""

_extension_array_result = """
>>> result = {result1}

.. plot::
    :context: close-figs
    :include-source: False

    >>> arr = sc.StairsArray({stair_list})
    >>> result = {result2}
    >>> fig, axes = plt.subplots(nrows=1, ncols=2,  figsize=(7,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ["result[0]", "result[1]"], result):
    ...     stair_instance.plot(ax=ax, arrows=True)
    ...     ax.set_title(title)
"""


def _make_extension_array_example(sym, rop=False):
    result = f"0.5 {sym} arr" if rop else f"arr {sym} 0.5"
    example = _extension_array_setup.format(
        var="arr", stair_list="[s1,s2]"
    ) + _extension_array_result.format(
        result1=result,
        sym=sym,
        stair_list="[s1,s2]",
        result2=result,
    )
    if not rop:
        example += _extension_array_setup.format(
            var="other", stair_list="[s4,s3]"
        ) + _extension_array_result.format(
            result1=f"arr {sym} other",
            sym=sym,
            stair_list="[s1,s2]",
            result2=f"arr {sym} sc.StairsArray([s4,s3])",
        )
    return example


_extension_array_binop = """
Binary operator for :class:`staircase.StairsArray` (and :class:`pandas.Series` with "Stairs" dtype).

Equivalent to {equiv}.

Parameters
----------
other : int, float, :class:`staircase.Stairs` or array-like of these
    If array-like must have same length as *self*.

Returns
-------
:class:`staircase.StairsArray`

See Also
--------
:meth:`staircase.Stairs.{funcstr}`

Examples
--------
{examples}
"""


def make_binop_docstring(funcstr):
    sym = {
        "ge": ">=",
        "gt": ">",
        "le": "<=",
        "lt": "<",
        "eq": "==",
        "ne": "!=",
        "add": "+",
        "subtract": "-",
        "multiply": "*",
        "divide": "/",
        "radd": "+",
        "rsubtract": "-",
        "rmultiply": "*",
        "rdivide": "/",
    }[funcstr]

    rop = funcstr in ("radd", "rsubtract", "rmultiply", "rdivide")
    if rop:
        equiv = f"*other* {sym} *self*"
    else:
        equiv = f"*self* {sym} *other*"
    return _extension_array_binop.format(
        funcstr=funcstr,
        equiv=equiv,
        examples=_make_extension_array_example(sym, rop),
    )


negate_docstring = """
Unary operator for :class:`staircase.StairsArray` (and :class:`pandas.Series` with "Stairs" dtype).

Equivalent to -*self*.

Returns
-------
:class:`staircase.StairsArray`

See Also
--------
:meth:`staircase.Stairs.negate`

Examples
--------

>>> arr = sc.StairsArray([s1, s2])

.. plot::
    :context: close-figs
    :include-source: False

    >>> arr = sc.StairsArray([s1, s2])
    >>> fig, axes = plt.subplots(nrows=1, ncols=2,  figsize=(7,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ["arr[0]", "arr[1]"], arr):
    ...     stair_instance.plot(ax=ax, arrows=True)
    ...     ax.set_title(title)

>>> result = -arr

.. plot::
    :context: close-figs
    :include-source: False

    >>> arr = sc.StairsArray([s1, s2])
    >>> result = -arr
    >>> fig, axes = plt.subplots(nrows=1, ncols=2,  figsize=(7,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ["result[0]", "result[1]"], result):
    ...     stair_instance.plot(ax=ax, arrows=True)
    ...     ax.set_title(title)
"""



import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from staircase.docstrings import examples
from staircase.util import _is_datetime_like
from staircase.util._decorators import Appender


# capable of single or vector
@Appender(examples.sample_example, join="\n", indents=1)
def sample(self, x, include_index=False):
    """
    Evaluates the value of the step function at one, or more, points.

    The function can be called using parentheses.  See example below.

    Parameters
    ----------
    x : int, float or vector data
        Values at which to evaluate the function
    include_index : bool, default False
        Indicates if the values returned should be a :class:`numpy.ndarray`, or in a :class:`pandas.Series`
        indexed by the values in *x*

    Returns
    -------
    float, :class:`numpy.ndarray`, :class:`pandas.Series`

    See Also
    --------
    Stairs.limit
    staircase.sample
    """
    side = "right" if self._closed == "left" else "left"
    values = limit(self, x, side)
    if include_index:
        if not is_list_like(x):
            x = [x]
        values = pd.Series(values, index=x)
    return values


@Appender(examples.limit_example, join="\n", indents=1)
def limit(self, x, side, include_index=False):
    """
    Evaluates the limit of the step function as it approaches one, or more, points.

    The results of this function should be considered as :math:`\\lim_{x \\to z^{-}} f(x)`
    or :math:`\\lim_{x \\to z^{+}} f(x)`, when side = 'left' or side = 'right' respectively. See
    :ref:`A note on interval endpoints<getting_started.interval_endpoints>` for an explanation.

    Parameters
    ----------
    x : int, float or vector data
        Values at which to evaluate the function
    side : {'left', 'right'}, default 'right'
        If points where step changes occur do not coincide with x then this parameter
        has no effect.  Where a step change occurs at a point given by x, this parameter
        determines if the step function is evaluated at the interval to the left, or the right.
    include_index : bool, default False
        Indicates if the values returned should be a :class:`numpy.ndarray`, or in a :class:`pandas.Series`
        indexed by the values in *x*

    Returns
    -------
    float, :class:`numpy.ndarray`, :class:`pandas.Series`

    See Also
    --------
    Stairs.sample
    staircase.sample
    """
    assert side in ("left", "right")
    passed_x = x
    if self._data is None:
        if pd.api.types.is_list_like(x):
            return self.initial_value * np.ones_like(x, dtype=float)
        else:
            return self.initial_value
    amended_values = np.append(
        self._get_values().values, [self.initial_value]
    )  # hack for -1 index value
    if pd.api.types.is_list_like(x) and _is_datetime_like(next(iter(x))):
        x = pd.Series(x).values  # faster, but also bug free in numpy
    elif _is_datetime_like(x):
        x = pd.Series([x]).values[0]
    values = amended_values[np.searchsorted(self._data.index.values, x, side=side) - 1]
    if include_index:
        values = pd.Series(values, index=passed_x)
    return values


def add_methods(cls):
    cls.sample = sample
    cls.limit = limit



import datetime
import warnings

import numpy as np
import pandas as pd
from pandas.api.types import (
    is_datetime64_any_dtype,
    is_list_like,
    is_number,
    is_numeric_dtype,
    is_timedelta64_dtype,
)

from staircase.constants import Inf, NegInf
from staircase.docstrings import examples
from staircase.util._decorators import Appender


def _check_args_dtypes(*vectors):
    approved_dtypes_checks = [
        is_datetime64_any_dtype,
        is_timedelta64_dtype,
        is_numeric_dtype,
    ]

    def _check_approved_dtype(vec):
        approved = any(map(lambda func: func(vec), approved_dtypes_checks))
        if not approved:
            warnings.warn(
                f"An argument supplied has dtype {vec.dtype}.  Only numerical, datetime-like, or timedelta dtypes have been tested.  Using other dtypes is considered experimental."
            )

    for vec in vectors:
        _check_approved_dtype(vec)


def _check_args_types(start, end):
    base_approved_types = [
        pd.Timedelta,
        pd.Timestamp,
        datetime.datetime,
        datetime.timedelta,
        np.datetime64,
        np.timedelta64,
    ]

    def _check_approved_type(arg, approved_types):
        approved = (
            arg is None
            or pd.api.types.is_number(arg)
            or isinstance(arg, approved_types)
        )
        if not approved:
            warnings.warn(
                f"An argument supplied for 'start' or 'end' has type '{type(arg)}'.  Only numerical, datetime-like, or timedelta types have been tested.  Using other types is considered experimental."
            )

    _check_approved_type(start, tuple(base_approved_types + [Inf]))
    _check_approved_type(end, tuple(base_approved_types + [NegInf]))


def _convert_to_series(vec):
    if not isinstance(vec, pd.Series):
        if vec is None or (isinstance(vec, (tuple, list)) and not len(vec)):
            vec = pd.Series(vec, dtype="float64")
        else:
            vec = pd.Series(vec)
    else:
        vec = vec.reset_index(drop=True)
    return vec


def _preprocess_layer_args(frame, start, end, value):
    if isinstance(start, NegInf):
        start = None
    if isinstance(end, Inf):
        end = None
    if frame is not None:
        start, end, value = _extract_from_frame(frame, start, end, value)
    if value is None:
        value = 1
    return start, end, value


def _extract_from_frame(frame=None, start=None, end=None, value=None):
    if frame is not None:
        frame = frame.copy()
    if isinstance(start, str):
        start = frame[start]
    if isinstance(end, str):
        end = frame[end]
    if isinstance(value, str):
        value = frame[value]

    return start, end, value


def _layer_scalar(self, start, end, value):
    _check_args_types(start, end)

    if start is not None and end is not None and start == end:
        return self

    if self._data is None:
        deltas = pd.Series(name="delta", dtype="float64")
    else:
        deltas = self._get_deltas()

    if start is None:
        self.initial_value += value
    else:
        if start in deltas.index:
            deltas.loc[start] += value
        else:
            deltas.loc[start] = value
        if deltas.loc[start] == 0:
            deltas.drop(start, inplace=True)

    if end is not None:
        if end in deltas.index:
            deltas.loc[end] -= value
        else:
            deltas.loc[end] = -value
        if deltas.loc[end] == 0:
            deltas.drop(end, inplace=True)
    if len(deltas) > 0:
        self._data = deltas.sort_index().to_frame()
        self._valid_deltas = True
    self._valid_values = False
    return self


@Appender(examples.layer_example, join="\n", indents=1)
def layer(self, start=None, end=None, value=None, frame=None):
    """
    Changes the values of the step function, in place, by 'layering' one or more intervals.

    The values of *start*, *end* and *value* parameters can be one of several types.
    If any of these is a string, then it is expected that *frame* is a :class:`pandas.DataFrame`
    and the string is the name of a column in the dataframe.

    If either *start*, or *end* evaluate to an array-like parameter then all of
    *start*, *end*, and *value* will be broadcast to the same length as the longest of these arrays.

    Parameters
    ----------
    start : scalar, array-like or string, default None
        Start point(s) of the interval(s).
        A value of None is interpreted as negative infinity.
    end : scalar, array-like or string, default None
        End points(s) of the interval(s).
        A value of None is interpreted as positive infinity.
    value : float, array-like or string, default None
        Value(s) of the interval(s).
        A value of None is equivalent to a value of 1.
    frame : :class:`pandas.DataFrame`, optional
        A dataframe containing named columns, whose names may appear as values
        for the other parameters.

    Returns
    -------
    :class:`Stairs`
        The current instance is returned to facilitate method chaining
    """
    if self._data is None and np.isnan(self.initial_value):
        return self
    self._clear_cache()
    start, end, value = _preprocess_layer_args(frame, start, end, value)
    if not any(list(map(is_list_like, (start, end, value)))):
        return _layer_scalar(self, start, end, value)
    value = np.array(value)
    assert not pd.isna(value).any(), "value parameter cannot contain null values"
    start = _convert_to_series(start)
    end = _convert_to_series(end)
    _check_args_dtypes(start, end)  # conversion to Series required before checking
    df = pd.concat([start, end], axis=1, ignore_index=True)
    start_series = pd.Series(value, index=df.iloc[:, 0])
    self.initial_value += start_series[start_series.index.isna()].sum()
    if self._data is None:
        to_concat = [
            start_series,
            pd.Series(-value, index=df.iloc[:, 1]),
        ]
    else:
        to_concat = [
            start_series,
            pd.Series(-value, index=df.iloc[:, 1]),
            self._get_deltas(),
        ]
    deltas = pd.concat(to_concat)

    self._data = deltas.groupby(deltas.index).sum().rename("delta").to_frame()
    self._valid_deltas = True
    self._valid_values = False
    self._remove_redundant_step_points()
    return self


def add_methods(cls):
    cls.layer = layer



import numbers
from collections.abc import Iterable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.api.extensions import (
    ExtensionArray,
    ExtensionDtype,
    register_extension_dtype,
)
from pandas.core.dtypes.inference import is_dict_like, is_list_like

from staircase.constants import inf
from staircase.core.arrays import docstrings
from staircase.core.stairs import Stairs
from staircase.core.stats.statistic import corr as _corr
from staircase.core.stats.statistic import cov as _cov
from staircase.util._decorators import Appender


@register_extension_dtype
class StairsDtype(ExtensionDtype):

    type = Stairs
    name = "Stairs"
    na_value = pd.NA

    @classmethod
    def construct_from_string(cls, string):
        if string == cls.name:
            return cls()
        else:
            raise TypeError("Cannot construct a '{}' from '{}'".format(cls, string))

    @classmethod
    def construct_array_type(cls):
        return StairsArray


def _isna(value):
    return pd.isna(value)


def _make_logical(func):
    def logical_func(x, axis=0):
        return np.where(
            np.logical_or.reduce(np.isnan(x), axis=axis),
            np.nan,
            func.reduce(x, axis=axis),
        )

    return logical_func


class StairsArray(ExtensionArray):

    _dtype = StairsDtype()
    ndim = 1

    def __init__(self, data):
        if isinstance(data, self.__class__):
            self.data = data.data
        elif isinstance(data, np.ndarray):
            if not data.ndim == 1:
                raise ValueError(
                    "'data' should be a 1-dimensional array of Stairs objects."
                )
            self.data = data
        elif is_dict_like(data):
            self.data = np.array([data[k] for k in data.keys()])
        elif isinstance(data, Stairs) or is_list_like(data):
            self.data = np.array(data, ndmin=1)
        else:
            raise TypeError("'data' should be array of Stairs objects.")

    @property
    def dtype(self):
        return self._dtype

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        if isinstance(idx, numbers.Integral):
            return self.data[idx]
        elif isinstance(idx, (Iterable, slice)):
            return StairsArray(self.data[idx])
        else:
            raise TypeError("Index type not supported", idx)

    def __setitem__(self, key, value):
        if isinstance(value, Stairs) or _isna(value):
            if _isna(value):
                # internally only use None as missing value indicator
                # but accept others
                value = None
            if isinstance(key, (slice, Iterable)):
                value_array = np.empty(1, dtype=object)
                value_array[:] = [value]
                self.data[key] = value_array
            else:
                self.data[key] = value

    def copy(self):
        return StairsArray(self.data.copy())

    def take(self, indices, allow_fill=False, fill_value=None):
        from pandas.api.extensions import take

        result = take(self.data, indices, allow_fill=allow_fill, fill_value=fill_value)
        if allow_fill and fill_value is None:
            result[pd.isna(result)] = None
        return StairsArray(result)

    def isna(self):
        return np.array([g is None for g in self.data], dtype="bool")

    def bool(self):
        result = np.array([bool(s) for s in self.data])
        return result

    @classmethod
    def _from_sequence(cls, scalars, dtype=None, copy=False):
        if isinstance(scalars, Stairs):
            scalars = [scalars]
        return StairsArray(scalars)

    @classmethod
    def _from_factorized(cls, values, original):
        return cls(values)

    @property
    def nbytes(self):
        return self._itemsize * len(self)

    @classmethod
    def _concat_same_type(cls, to_concat):
        return cls(np.concatenate([array.data for array in to_concat]))

    @Appender(docstrings.make_docstring("array", "mean"), join="\n", indents=1)
    def mean(self):
        return self.agg(np.mean)

    @Appender(docstrings.make_docstring("array", "median"), join="\n", indents=1)
    def median(self):
        return self.agg(np.median)

    @Appender(docstrings.make_docstring("array", "sum"), join="\n", indents=1)
    def sum(self):
        return self.agg(np.sum)

    @Appender(docstrings.make_docstring("array", "min"), join="\n", indents=1)
    def min(self):
        return self.agg(np.min)

    @Appender(docstrings.make_docstring("array", "max"), join="\n", indents=1)
    def max(self):
        return self.agg(np.max)

    @Appender(docstrings.make_docstring("array", "agg"), join="\n", indents=1)
    def agg(self, func):
        index = pd.Index(
            np.unique(
                np.concatenate(
                    [
                        s.step_changes.index  # using .step_changes.index instead of .step_points to retain timezones
                        for s in (sf for sf in self.data if sf.number_of_steps)
                    ]
                )
            )
        )
        new_values = func([s.limit(index, "right") for s in self.data], axis=0)
        if new_values.dtype == "bool":
            new_values = new_values.astype(int)

        return Stairs._new(
            initial_value=func([s.initial_value for s in self.data]),
            data=pd.Series(
                new_values,
                index=index,
                name="value",
            ).to_frame(),
            closed=self.data[0].closed,
        )._remove_redundant_step_points()

    @Appender(docstrings.make_docstring("array", "sample"), join="\n", indents=1)
    def sample(self, x):
        array = pd.Series(self.data)
        return array.apply(Stairs.sample, x=x, include_index=True)

    @Appender(docstrings.make_docstring("array", "limit"), join="\n", indents=1)
    def limit(self, x, side="right"):
        array = pd.Series(self.data)
        return array.apply(Stairs.limit, x=x, side=side, include_index=True)

    @Appender(docstrings.make_docstring("array", "logical_or"), join="\n", indents=1)
    def logical_or(self):
        return self.agg(_make_logical(np.logical_or))

    @Appender(docstrings.make_docstring("array", "logical_and"), join="\n", indents=1)
    def logical_and(self):
        return self.agg(_make_logical(np.logical_and))

    @Appender(docstrings.make_docstring("array", "plot"), join="\n", indents=1)
    def plot(self, ax=None, labels=None, **kwargs):
        if ax is None:
            _, ax = plt.subplots()
        if labels is None:
            labels = range(len(self))
        elif len(labels) != len(self):
            raise ValueError(
                f"Number of labels supplied: {len(labels)}  Required: {len(self)}."
            )
        for s, l in zip(self, labels):
            s.plot(ax=ax, label=l, style="step", **kwargs)
        return ax

    def _reduce(self, name: str, *, skipna: bool = True, **kwargs):
        return {
            "sum": self.sum,
            "mean": self.mean,
            "min": self.min,
            "max": self.max,
            "median": self.median,
            "corr": self.corr,
            "agg": self.agg,
        }[name]()

    @Appender(docstrings.negate_docstring, join="\n", indents=1)
    def negate(self):
        return StairsArray(-self.data)

    __neg__ = negate


def _make_binary_func(func_str):

    stairs_func = getattr(Stairs, func_str)
    docstring = docstrings.make_binop_docstring(funcstr)

    @Appender(docstring, join="\n", indents=1)
    def func(self, other):
        if isinstance(other, (int, float, Stairs)):
            result = StairsArray([stairs_func(s, other) for s in self.data])
        elif is_list_like(other):
            if len(other) != len(self):
                return ValueError("Arrays have different lengths.")
            result = StairsArray(
                [stairs_func(s1, s2) for s1, s2 in zip(self.data, other.data)]
            )
        else:
            return NotImplemented
        return result

    return func


for funcstr in (
    "ge",
    "gt",
    "le",
    "lt",
    "eq",
    "ne",
    "add",
    "subtract",
    "multiply",
    "divide",
    "radd",
    "rsubtract",
    "rmultiply",
    "rdivide",
):
    func = _make_binary_func(funcstr)
    setattr(StairsArray, funcstr, func)
    magic = {
        "subtract": "sub",
        "multiply": "mul",
        "divide": "truediv",
        "rsubtract": "rsub",
        "rmultiply": "rmul",
        "rdivide": "rtruediv",
    }.get(funcstr, funcstr)
    setattr(StairsArray, f"__{magic}__", func)


def _make_corr_cov_func(docstring, stairs_method, assume_ones_diagonal):
    @Appender(docstring, join="\n", indents=1)
    def func(self, where=(-inf, inf)):
        size = len(self.data)
        vals = np.ones(shape=(size, size))
        for i in range(size):
            for j in range(i + assume_ones_diagonal, size):
                vals[i, j] = stairs_method(self.data[i], self.data[j], where=where)
                vals[j, i] = vals[i, j]
        return vals

    return func


StairsArray.corr = _make_corr_cov_func(
    docstrings.make_docstring("array", "corr"), _corr, assume_ones_diagonal=True
)
StairsArray.cov = _make_corr_cov_func(
    docstrings.make_docstring("array", "cov"), _cov, assume_ones_diagonal=False
)



from staircase.core.stairs import Stairs, _add_operations  # isort: skip

_add_operations()

from staircase.constants import inf
from staircase.core.arrays import (
    agg,
    corr,
    cov,
    limit,
    logical_and,
    logical_or,
    max,
    mean,
    median,
    min,
    plot,
    sample,
    sum,
)
from staircase.core.arrays.extension import StairsArray
from staircase.core.slicing import StairsSlicer
from staircase.core.stats.distribution import Dist
from staircase.test_data import make_test_data


def get_version():
    def get_version_post_py38():
        from importlib.metadata import version  # type: ignore

        return version(__name__)

    def get_version_pre_py38():
        from pkg_resources import get_distribution

        return get_distribution(__name__).version

    def default_version():
        return "unknown"

    for func in (get_version_post_py38, get_version_pre_py38, default_version):
        try:
            return func()
        except Exception:
            pass


__version__ = get_version()



"""
General exceptions
"""


class ClosedMismatchError(ValueError):
    def __init__(self, stairs1, stairs2):
        """
        Create a `ClosedMismatchError` indicating that `stairs1` and
        `stairs2` were supposed to have the same `clased` values but didn't.
        """
        super().__init__(
            f"`closed` values must be same but were {stairs1._closed} and {stairs2._closed}"
        )



import numpy as np
import pandas as pd


def make_test_data(dates=True, positive_only=True, groups=(), seed=None):
    """
    Creates interval data for use with staircase.

    The result will be a :class:`pandas.DataFrame` with columns "start", "end", "value", and optionally "group".

    Parameters
    ----------
    dates : bool, default True
        Indicates whether to return data associated with a datetime domain or numerical.
        If dates is true the data will be confined to the year 2021.
        If dates is false the data will be confined to the interval [0, 100]
    positive_only : bool, default True
        If True then *value* column will only consist of positive values
    groups : array-like of strings, optional
        If specified will create data for each string specified in *groups*.  A column
        named "group" will be added to identify the subsets of data.
    seed : int, optional
        If specified will seed a random number generator to facilitate reproducability.

    Returns
    -------
    :class:`pandas.DataFrame`

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> df = sc.make_test_data()
        >>> sc.Stairs(df, "start", "end").plot()

    .. plot::
        :context: close-figs

        >>> df = sc.make_test_data(dates=False, positive_only=False, seed=42)
        >>> sc.Stairs(df, "start", "end", "value").plot()

    .. plot::
        :context: close-figs

        >>> df = sc.make_test_data(groups=["A", "B", "C"])
        >>> stairs = df.groupby("group").apply(sc.Stairs, "start", "end")
        >>> ax = stairs["A"].plot(label="A")
        >>> stairs["B"].plot(ax, label="B")
        >>> stairs["C"].plot(ax, label="C")
        >>> ax.legend()

    """
    if len(groups) == 0:
        return _make_test_data(dates=dates, positive_only=positive_only, seed=seed)

    dfs = []

    for group in groups:
        df = _make_test_data(dates=dates, positive_only=positive_only, seed=seed)
        dfs.append(df.assign(group=group))
        if seed is not None:
            seed += 1

    return pd.concat(dfs)


def _get_random_functions(seed=None):
    major, minor, *_ = np.__version__.split(".")

    if f"{major}.{minor}" in ("1.14", "1.15", "1.16"):
        rs = np.random.RandomState(seed)
        return rs.randint, rs.uniform

    rng = np.random.default_rng(seed)
    return rng.integers, rng.random


def _make_test_data(dates=True, positive_only=True, seed=None):

    random_integers, random_floats = _get_random_functions(seed)

    n_intervals = random_integers(100, 1000)
    smallest_interval = random_integers(1, 6)
    largest_interval = random_integers(15, 20)
    interval_lengths = (
        random_floats(size=n_intervals) * (largest_interval - smallest_interval)
        + smallest_interval
    )

    start = random_floats(size=n_intervals) * 120 - 20
    end = start + interval_lengths

    if dates:
        start = pd.Timestamp("2021") + pd.TimedeltaIndex(
            start * 365 / 100 * pd.Timedelta(1, "day")
        )
        end = pd.Timestamp("2021") + pd.TimedeltaIndex(
            end * 365 / 100 * pd.Timedelta(1, "day")
        )

    value = random_integers(-5, 6, n_intervals)
    if positive_only:
        value = value + 6

    data = pd.DataFrame(
        {
            "start": start,
            "end": end,
            "value": value,
        }
    )

    if dates:
        data["start"] = data["start"].dt.floor("min")
        data["end"] = data["end"].dt.floor("min")
        data = data.loc[(data["end"] > "2021") & (data["start"] < "2022")]
        data.loc[data["start"] < "2021", "start"] = pd.NaT
        data.loc[data["end"] >= "2022", "end"] = pd.NaT
    else:
        data = data.loc[(data["end"] > 0) & (data["start"] < 100)]
        data.loc[data["start"] < 0, "start"] = np.nan
        data.loc[data["end"] >= 100, "end"] = np.nan

    return data.sort_values(["end", "start"]).reset_index(drop=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['staircase',
 'staircase.core',
 'staircase.core.arrays',
 'staircase.core.exceptions',
 'staircase.core.ops',
 'staircase.core.stats',
 'staircase.docstrings',
 'staircase.io',
 'staircase.plotting',
 'staircase.util']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=2', 'pytz']

extras_require = \
{':python_version >= "3.10" and python_version < "4.0"': ['pandas>=1.3.4,<2.0.0',
                                                          'numpy>=1.21.2,<2.0.0'],
 ':python_version >= "3.7" and python_version < "3.10"': ['pandas>=1,<2',
                                                          'numpy>=1.15,<2.0']}

setup_kwargs = {
    'name': 'staircase',
    'version': '2.5.0',
    'description': 'A data analysis package based on modelling and manipulation of mathematical step functions. Strongly aligned with pandas.',
    'long_description': '<p align="center"><a href="https://github.com/staircase-dev/staircase"><img src="https://github.com/staircase-dev/staircase/blob/master/docs/img/staircase2.png?raw=true" title="staircase logo" alt="staircase logo"></a></p>\n\n\n<p align="center">\n\t<a href="https://pepy.tech/project/staircase/" alt="PyPI downloads">\n        <img src="https://pepy.tech/badge/staircase" /></a>\n    <a href="https://www.python.org/" alt="Python version">\n        <img src="https://img.shields.io/pypi/pyversions/staircase" /></a>\n    <a href="https://pypi.org/project/staircase/" alt="PyPI version">\n        <img src="https://img.shields.io/pypi/v/staircase" /></a>\n    <a href="https://anaconda.org/conda-forge/staircase" alt="Conda Forge version">\n        <img src="https://anaconda.org/conda-forge/staircase/badges/version.svg?branch=master&kill_cache=1" /></a>\n    <a href="https://github.com/staircase-dev/staircase/blob/master/LICENSE" alt="License">\n        <img src="http://img.shields.io/:license-mit-blue.svg?style=flat-square"></a>\n</p>\n<p align="center">\n\t<a href="https://github.com/staircase-dev/staircase/actions/workflows/ci.yml" alt"Github CI">\n\t\t<img src="https://github.com/staircase-dev/staircase/actions/workflows/ci.yml/badge.svg"/></a>\n    <a href="https://www.staircase.dev/en/latest/" alt="Documentation">\n        <img src="https://readthedocs.org/projects/railing/badge/?version=latest" /></a>\n\t<a href="https://app.codacy.com/gh/staircase-dev/staircase/dashboard" alt="Codacy Grade">\n        <img src="https://app.codacy.com/project/badge/Grade/845ecfb2fd6748cc87a66f9a97cd9492" /></a>\t\n\t<a href="https://app.codecov.io/gh/staircase-dev/staircase"  alt="Codecov coverage">\n\t\t<img src="https://codecov.io/gh/staircase-dev/staircase/branch/master/graph/badge.svg"/></a>\n</p>\n\nThe staircase package enables data analysis through mathematical step functions. Step functions can be used to represent continuous time series - think changes in state over time, queue size over time, utilisation over time, success rates over time etc.\n\nThe package is built upon `numpy` and `pandas`, with a deliberate, stylistic alignment to the latter in order to integrate seamlessly into the [pandas ecosystem](https://pandas.pydata.org/docs/ecosystem.html).\n\nThe staircase package makes converting raw, temporal data into time series easy and readable. Furthermore there is a rich variety of [arithmetic operations](https://www.staircase.dev/en/latest/reference/Stairs.html#arithmetic-operators), [relational operations](https://www.staircase.dev/en/latest/reference/Stairs.html#relational-operators), [logical operations](https://www.staircase.dev/en/latest/reference/Stairs.html#logical-operators), [statistical operations](https://www.staircase.dev/en/latest/reference/Stairs.html#statistical-operators), to enable analysis, in addition to functions for [univariate analysis](https://www.staircase.dev/en/latest/reference/Stairs.html#summary-statistics), [aggregations](https://www.staircase.dev/en/latest/reference/arrays.html#aggregation) and compatibility with datetimes.\n\n**New in 2022:** staircase now provides support for [pandas extension arrays](https://pandas.pydata.org/docs/ecosystem.html#extension-data-types) and a [Series accessor](https://www.staircase.dev/en/latest/user_guide/arraymethods.html).\n\n\n## An example\n\nIn this example, we consider data corresponding to site views for a website in October 2021.  The start and end times have been logged for each session, in addition to one of three countries codes (AU, UK, US).  These times are recorded with `pandas.Timestamp` and any time which falls outside of October is logged as `NAT`.\n\n\n```python\n>>> data\n                       start                   end   country\n0                        NaT   2021-10-01 00:00:50        AU\n1                        NaT   2021-10-01 00:07:45        AU\n2                        NaT   2021-10-01 00:05:58        AU\n3                        NaT   2021-10-01 00:08:48        AU\n4                        NaT   2021-10-01 00:05:26        AU\n...                      ...                   ...       ...\n425728   2021-10-31 23:57:16                   NaT        US\n425729   2021-10-31 23:57:25                   NaT        US\n425730   2021-10-31 23:58:59                   NaT        US\n425731   2021-10-31 23:59:45                   NaT        US\n425732   2021-10-31 23:59:59                   NaT        US\n```\n\nNote that the number of users viewing the site over time can be modelled as a step function.  The value of the function increases by 1 every time a user arrives at the site, and decreases by 1 every time a user leaves the site.  This step function can be thought of as the sum of three step functions: AU users + UK users + US users.  Creating a step function for AU users, for example, is simple.  To achieve it we use the *[Stairs](https://www.staircase.dev/en/latest/reference/Stairs.html)* class, which represents a step function:\n\n\n```python\n>>> import staircase as sc\n\n>>> views_AU = sc.Stairs(data.query("country == \'AU\'"), "start", "end")\n>>> views_AU\n<staircase.Stairs, id=1609972469384>\n```\n\nWe can visualise the function with the plot function:\n```python\n>>> views_AU.plot()\n```\n\n<p align="left"><img src="https://github.com/staircase-dev/staircase/blob/master/docs/img/AU_views.png?raw=true" title="AU views example" alt="AU views example"></p>\n\nRather than creating a separate variable for each country, we can create a `pandas.Series` to hold a step function for each country.  We can even give this Series a "Stairs" type.\n\n```python\n>>> october = (pd.Timestamp("2021-10"), pd.Timestamp("2021-11"))\n>>> series_stepfunctions = (\n...     data.groupby("country")\n...     .apply(sc.Stairs, "start", "end")\n...     .apply(sc.Stairs.clip, october)  # set step functions to be undefined outside of October\n...     .astype("Stairs")\n... )\n>>> series_stepfunctions\ncountry\nAU    <staircase.Stairs, id=2516367680328>\nUK    <staircase.Stairs, id=2516362550664>\nUS    <staircase.Stairs, id=2516363585928>\ndtype: Stairs\n```\n\nThe plotting backend to `staircase` is provided by `matplotlib`.\n\n```python\n>>> import matplotlib.pyplot as plt\n>>> _, ax = plt.subplots(figsize=(15,4))\n>>> series_stepfunctions.sc.plot(ax, alpha=0.7)\n>>> ax.legend()\n```\n<p align="left"><img src="https://github.com/staircase-dev/staircase/blob/master/docs/img/all_views.png?raw=true" title="all views example" alt="all views example"></p>\n\nNow plotting step functions is useful, but the real fun starts when we go beyond this:\n\n<p align="left"><img src="https://github.com/staircase-dev/staircase/blob/master/docs/img/staircase_analysis.gif?raw=true" title="staircase analysis examples" alt="staircase analysis examples"></p>\n\n\n## Installation\n\nstaircase can be installed from PyPI:\n\n```bash\npython -m pip install staircase\n```\n\nor also with conda:\n\n```bash\nconda install -c conda-forge staircase\n```\n\n## Documentation\nThe complete guide to using staircase can be found at [staircase.dev](https://www.staircase.dev)\n\n## Contributing\nThere are many ways in which contributions can be made - the first and foremost being *using staircase and giving feedback*.\n\nBug reports, feature requests and ideas can be submitted via the [Github issue tracker](https://github.com/staircase-dev/staircase/issues).\n\nAdditionally, bug fixes. enhancements, and improvements to the code and documentation are also appreciated and can be done via pull requests.\nTake a look at the current issues and if there is one you would like to work on please leave a comment to that effect.\n\nSee this [beginner\'s guide to contributing](https://github.com/firstcontributions/first-contributions), or [Pandas\' guide to contributing](https://pandas.pydata.org/pandas-docs/stable/development/contributing.html), to learn more about the process.\n\n\n## Versioning\n\nWe use [SemVer](http://semver.org/) for versioning. For the versions available, see the [tags on this repository](https://github.com/staircase-dev/staircase/tags).  It is highly recommended to use staircase 2.*, for both performance and additional features.\n\n\n## License\n\nThis project is licensed under the MIT License - see the [LICENSE](https://github.com/staircase-dev/staircase/blob/master/LICENSE) file for details\n\n## Acknowledgments\n\nThe seeds of *staircase* began developing at the Hunter Valley Coal Chain Coordinator, where it finds strong application in analysing simulated data.  Thanks for the support!\n',
    'author': 'Riley Clement',
    'author_email': 'venaturum@gmail.com',
    'maintainer': 'Riley Clement',
    'maintainer_email': 'venaturum@gmail.com',
    'url': 'https://staircase.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import pandas as pd
from pandas.api.extensions import register_series_accessor

from staircase.constants import inf
from staircase.core.arrays import docstrings
from staircase.core.arrays.extension import StairsDtype
from staircase.util._decorators import Appender


@register_series_accessor("sc")
class StairsAccessor:
    def __init__(self, pandas_obj):
        if not isinstance(pandas_obj.dtype, StairsDtype):
            raise AttributeError(
                'sc accessor only valid for Series with Stairs dtype.  Convert using .astype("Stairs").'
            )
        self._obj = pandas_obj

    @Appender(docstrings.make_docstring("accessor", "corr"), join="\n", indents=1)
    def corr(self, where=(-inf, inf)):
        return pd.DataFrame(
            self._obj.values.corr(where),
            index=self._obj.index,
            columns=self._obj.index,
        )

    @Appender(docstrings.make_docstring("accessor", "cov"), join="\n", indents=1)
    def cov(self, where=(-inf, inf)):
        return pd.DataFrame(
            self._obj.values.cov(where),
            index=self._obj.index,
            columns=self._obj.index,
        )

    @Appender(docstrings.make_docstring("accessor", "logical_or"), join="\n", indents=1)
    def logical_or(self):
        return self._obj.values.logical_or()

    @Appender(
        docstrings.make_docstring("accessor", "logical_and"), join="\n", indents=1
    )
    def logical_and(self):
        return self._obj.values.logical_and()

    @Appender(docstrings.make_docstring("accessor", "sample"), join="\n", indents=1)
    def sample(self, x):
        result = self._obj.values.sample(x)
        result.index = self._obj.index
        return result

    @Appender(docstrings.make_docstring("accessor", "limit"), join="\n", indents=1)
    def limit(self, x, side="right"):
        result = self._obj.values.limit(x, side)
        result.index = self._obj.index
        return result

    @Appender(docstrings.make_docstring("accessor", "plot"), join="\n", indents=1)
    def plot(self, ax=None, **kwargs):
        labels = self._obj.index
        return self._obj.values.plot(ax, labels, **kwargs)



from staircase.plotting import docstrings
from staircase.plotting.matplotlib import plot as plot_matplotlib
from staircase.util._decorators import Appender


class PlotAccessor:
    def __init__(self, stairs):
        self._stairs = stairs

    def _dist(self):
        return self._stairs.dist

    @Appender(docstrings.matplotlib_docstring, join="\n", indents=1)
    def __call__(self, *args, backend="matplotlib", **kwargs):
        # register_matplotlib_converters()
        # if self._stairs._data is None:
        #    return
        if backend == "matplotlib":
            return plot_matplotlib(self._stairs, *args, **kwargs)
        else:
            raise ValueError(f"Plotting backend {backend} not defined")

    def ecdf(self, backend="matplotlib", **kwargs):
        if backend == "matplotlib":
            return plot_matplotlib(self._dist().ecdf, **kwargs)
        else:
            raise ValueError(f"Plotting backend {backend} not defined")
